#include "page_queue_partitioned.h"



/* Discard all queued flexpages. */

void PageQueuePartitioned::close(Memory *memory)
{
    discard(_available, memory);
    discard(_issued, memory);
}

/* Keep waiting for a potential queue non-empty condition.
   Then, attempt to pop an entry from the queue. */

void PageQueuePartitioned::pop(PageOwner **owner, Flexpage **flexpage)
{
    std::unique_lock<std::mutex> guard(_lock);
    QueueEntry entry;

    while (1)
    {
        if (_pop(&entry))
        {
            *owner = entry.owner;
            *flexpage = entry.flexpage;
            return;
        }
        else
            _counter.wait(guard);
    }
}

/* Check the available pages queue for entries, returning false if no entries
   are available, returning true and providing the details if an entry can be
   removed from the front of the queue. */

bool PageQueuePartitioned::_pop(QueueEntry *entry)
{
    if (_available.empty())
        return false;

    *entry = _available.front();
    _available.pop_front();

    return true;
}

/* Push an entry for the given owner and flexpage to the appropriate queue. */

void PageQueuePartitioned::push(PageOwner *owner, Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Record the entry and a position reference for the flexpage. */

    Queue *queue;
    Positions *positions = NULL;

    if (owner == NULL)
        queue = &_available;
    else
    {
        queue = &_issued;
        positions = &_positions;
    }

    queue->push_back((QueueEntry) {flexpage, owner});

    if (positions != NULL)
    {
        Queue::iterator last = queue->end();
        last--;
        positions->insert(Position(flexpage, last));
    }

    _counter.notify_one();
}

/* Push an entry to the front of the appropriate queue. */

void PageQueuePartitioned::push_front(PageOwner *owner, Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    Queue *queue;
    Positions *positions = NULL;

    if (owner == NULL)
        queue = &_available;
    else
    {
        queue = &_issued;
        positions = &_positions;
    }

    queue->push_back((QueueEntry) {flexpage, owner});

    if (positions != NULL)
        positions->insert(Position(flexpage, queue->begin()));

    _counter.notify_one();
}

/* Remove an entry for the given owner and flexpage from the queue. */

bool PageQueuePartitioned::remove(PageOwner *owner, Flexpage *flexpage)
{
    return PageQueue::remove(_issued, _positions, owner, flexpage);
}

// vim: tabstop=4 expandtab shiftwidth=4
